;;;; "Advanced Choreography" -- this is the final video for this project.

(ns com.aurellem.run.adv-choreo
  (:use (com.aurellem.gb saves gb-driver util constants
                         items vbm characters money
                         rlm-assembly))
  (:use (com.aurellem.run util music title save-corruption
                          bootstrap-0 bootstrap-1 image
                          ram-display final-cut basic-choreo))
  (:require clojure.string)
  (:import java.awt.image.BufferedImage)
  (:import (javax.imageio ImageWriteParam IIOImage ImageIO))
  (:import [com.aurellem.gb.gb_driver SaveState])
  (:import java.io.File))



;; Use the gameboy's screen to display the new programming
;; instead of a side window. This will make it look much
;; cooler and create a terminal-like effect as the game is
;; being reprogramed. To do this, use a fixed data entry
;; region in ram, and run a program that translates this
;; region into the screen. Every time this data entry region
;; is full, run a program that copies the data to the
;; appropriate region in memory. This will cost ~15 seconds
;; at the beginning to set up, and then should have minimal
;; overhead (~5%) for the rest of the data transfer, but
;; will have a good psychological effect for the viewer
;; since he can see that something is actually happening in
;; the game.


;; Symbol size and type.

;; use fonts from zophar's domain:
;; http://www.zophar.net/utilities/fonts/8x8-font-archive.html

;; Green font on black background for matrix look.


(defn program-data [base-address]
  (let [image-program
        (display-image-kernel
         base-address
         pinkie-pie-mark)

        music-base-address (+ (count image-program) base-address)

        initial-music-data
        (midi-bytes pony-csv 0 0 0 0)
        
        data-lengths
        (map (comp count :data)
                   [(:kernel  initial-music-data)
                    (:voice-1 initial-music-data)
                    (:voice-2 initial-music-data)]);; noise not needed
        addresses
        (map (partial + music-base-address) (reductions + 0 data-lengths))

        final-music-data
        (apply (partial midi-bytes pony-csv) addresses)

        music-program
        (concat
         (:data (:kernel  final-music-data))
         (:data (:voice-1 final-music-data))
         (:data (:voice-2 final-music-data))
         (:data (:noise   final-music-data)))]

    (concat
     image-program ;; image program falls through to music program
     music-program)))




(def glyphs
  "The sixteen 8x8 glyphs which make up the \"terminal\" font."
  (mapv #(ImageIO/read
          (File. user-home (str "proj/vba-clojure/font/" % ".png")))
          ["0" "1" "2" "3" "4" "5" "6" "7" "8" "9" "A" "B" "C" "D" "E" "F"]))

(defn glyphs-init-program
  [start-address]
  (let [zero-glyph (image->gb-image (glyphs 0))
        
        ;; write same pallet information to all pallettes
        A (flatten
           [(write-byte LCD-control-register 0x00)
            (set-palettes bg-palette-select bg-palette-data
                          (repeat 8 (first (:palettes zero-glyph))))
            (select-LCD-bank 0)
            (write-byte SCX-register 0)
            (write-byte SCY-register 0)])
        B (flatten
           [(write-data 
             (+ start-address (count A))
             character-data-address
             (flatten
              (map (comp gb-tile->bytes first :tiles image->gb-image)
                   glyphs)))])]
    (concat A B)))


(defn glyph-display-program
  [start-address
   delay-count
   total-glyph-count]
  (glyphs-init-program start-address)  ;; ONLY  for testing!
  )


(defn glyph-bootstrap-program
  [start-address delay-count total-glyph-count]
  (let [init [0xAF 0x4F 0x47] ;; 0->A; 0->C; 0->B
        header (concat (frame-metronome) (read-user-input))

        glyph-display (glyph-display-program
                       (+ (count init) (count header)
                          start-address)
                       0 0) ;; ONLY FOR TESTING

        state-machine-start-address
        (+ start-address (count init) (count header) (count glyph-display))
        state-machine
        (bootstrap-state-machine state-machine-start-address)
        
        return-to-header
        (flatten 
         [0x18
          (->signed-8-bit
           (- (count init)
              2 ;; this command length
              3 ;; I have no idea why we need a 3 here
              ;; need to investigate.
              (count glyph-display)
              (count header)
              (count state-machine)))])]
    
    (concat init glyph-display header state-machine return-to-header)))




(def main-program-base-address 0xC000)



;; RLM want to transfer control here --- this is where I left off.
(defn begin-glyph-bootstrap
  ([] (begin-glyph-bootstrap (relocate-main-bootstrap)))
  ([script]
     (let [glyph-program (glyph-bootstrap-program
                          0xDF00 0 0)


           

           
(defn write-all-program-data
  ([] (write-all-program-data (relocate-main-bootstrap)))
  ([script]
     (let [base-address main-program-base-address]
       (->> script
            (write-RAM 0xFF1A [0 0 0]) ;; silence remnant music
            (write-RAM base-address (program-data base-address))))))

(defn activate-program
  ([] (activate-program (write-all-program-data)))
  ([script]
     (->> script
          (transfer-control main-program-base-address)
          (do-nothing 1800))))


;; possible screen writing programs

;; (program needs to stop executing at some point)
;; maybe have total length counter or something?

;; automatic counter that reads from program-start and clears the
;; screen every 360 (* 18 20) gliphs

;; advantages -- very simple and low bandwidth
;; disadvantages -- hard to align counter

;; implementation -- refactor main-bootstrap-program to provide a
;; state-machine code-section which can be recombined into another
;; program.
